# 同步复习题

> 原文：<https://github.com/angrave/SystemProgramming/wiki/Synchronization-Review-Questions>

## 话题

*   原子操作
*   关键部分
*   生产者消费者问题
*   使用条件变量
*   使用计数信号量
*   实施障碍
*   实现环形缓冲区
*   使用 pthread_mutex
*   实施生产者消费者
*   分析多线程编码

## 问题

*   什么是原子操作？
*   为什么以下不能在并行代码中工作

```c
//In the global section
size_t a;
//In pthread function
for(int i = 0; i < 100000000; i++) a++;
```

这会吗？

```c
//In the global section
atomic_size_t a;
//In pthread function
for(int i = 0; i < 100000000; i++) atomic_fetch_add(a, 1);
```

*   原子操作有哪些缺点？什么会更快：保持局部变量或许多原子操作？
*   什么是关键部分？
*   一旦确定了一个关键部分，确保一次只有一个线程在该部分中的一种方法是什么？
*   在此确定关键部分

```c
struct linked_list;
struct node;
void add_linked_list(linked_list *ll, void* elem){
    node* packaged = new_node(elem);
    if(ll->head){
         ll->head = 
    }else{
         packaged->next = ll->head;
         ll->head = packaged;
         ll->size++;
    }

}

void* pop_elem(linked_list *ll, size_t index){
    if(index >= ll->size) return NULL;

    node *i, *prev;
    for(i = ll->head; i && index; i = i->next, index--){
        prev = i;
    }

    //i points to the element we need to pop, prev before
    if(prev->next) prev->next = prev->next->next;
    ll->size--;
    void* elem = i->elem;
    destroy_node(i);
    return elem;
}
```

你有多紧张关键部分？

*   什么是生产者消费者问题？在上一节中如何使用上述生产者消费者问题？生产者消费者问题与读者作家问题有什么关系？
*   什么是条件变量？为什么在`while`循环上使用一个有优势？
*   为什么这段代码很危险？

```c
if(not_ready){
     pthread_cond_wait(&cv, &mtx);
}
```

*   什么是计数信号量？给我一个饼干罐/披萨盒/限量食品的类比。
*   什么是线程障碍？
*   使用计数信号量来实现屏障。

*   编写生产者/消费者队列，生产者消费者栈怎么样？

*   给我一个带条件变量的读写器锁的实现，用你需要的任何东西制作一个结构，它只需要能够支持以下函数

```c
void reader_lock(rw_lock_t* lck);
void writer_lock(rw_lock_t* lck);
void reader_unlock(rw_lock_t* lck);
void writer_unlock(rw_lock_t* lck);
```

唯一的规范是在`reader_lock`和`reader_unlock`之间，没有编写者可以写。在写入器锁之间，一次只能写一个作者。

*   编写代码以使用仅三个计数信号量来实现生产者使用者。假设可以有多个线程调用 enqueue 和 dequeue。确定每个信号量的初始值。
*   编写代码以使用条件变量和互斥锁实现生产者使用者。假设可以有多个线程调用 enqueue 和 dequeue。
*   使用 CV 实现 add（unsigned int）和 subtract（unsigned int）阻塞函数，这些函数永远不会允许全局值大于 100。
*   使用 CV 为 15 个线程实现屏障。
*   以下有多少陈述是正确的？
    *   可以有多个活跃的读者
    *   可以有多个活动作者
    *   当有活动的写入器时，活动读取器的数量必须为零
    *   如果有活动的阅读器，则活动写入器的数量必须为零
    *   作者必须等到当前活跃的读者完成
*   Todo：分析多线程代码片段