# 信号，第 4 部分：信号

> 原文：<https://github.com/angrave/SystemProgramming/wiki/Signals%2C-Part-4%3A-Sigaction>

## 我如何以及为何使用`sigaction`？

您应该使用`sigaction`而不是`signal`，因为它具有更好的定义语义。 `signal`在不同的操作系统上做了不同的事情，**坏** `sigaction`更便携，如果需要更好地为线程定义。

要改变过程的“信号处理” - 即当信号传递到您的过程时会发生什么 - 使用`sigaction`

您可以使用系统调用`sigaction`来设置信号的当前处理程序，或者读取特定信号的当前信号处理程序。

```c
int sigaction(int signum, const struct sigaction *act, struct sigaction *oldact);
```

sigaction 结构包括两个回调函数（我们只会查看'handler'版本），一个信号掩码和一个 flags 字段 -

```c
struct sigaction {
               void     (*sa_handler)(int);
               void     (*sa_sigaction)(int, siginfo_t *, void *);
               sigset_t   sa_mask;
               int        sa_flags;
}; 
```

## 如何将`signal`调用转换为等效的`sigaction`调用？

假设您为警报信号安装了信号处理程序，

```c
signal(SIGALRM, myhandler);
```

等效的`sigaction`代码是：

```c
struct sigaction sa; 
sa.sa_handler = myhandler;
sigemptyset(&sa.sa_mask);
sa.sa_flags = 0; 
sigaction(SIGALRM, &sa, NULL)
```

但是，我们通常也可以设置掩码和标志字段。掩码是在信号处理程序执行期间使用的临时信号掩码。 SA_RESTART 标志将自动重启一些（但不是全部）系统调用，否则这些调用将提前返回（带有 EINTR 错误）。后者意味着我们可以稍微简化代码的其余部分，因为可能不再需要重启循环。

```c
sigfillset(&sa.sa_mask);
sa.sa_flags = SA_RESTART; /* Restart functions if  interrupted by handler */     
```

## 我如何使用 sigwait？

Sigwait 可用于一次读取一个待处理信号。 `sigwait`用于同​​步等待信号，而不是在回调中处理它们。下面显示了多线程程序中 sigwait 的典型用法。请注意，首先设置线程信号掩码（并且将由新线程继承）。这可以防止信号被 _ 传递 _，因此它们将保持挂起状态，直到调用 sigwait。还要注意 sigwait 使用相同的 set sigset_t 变量 - 除了设置阻塞信号集之外，它被用作 sigwait 可以捕获和返回的信号集。

编写自定义信号处理线程（例如下面的示例）而不是回调函数的一个优点是，您现在可以使用更多的 C 库和系统函数，否则这些函数无法在信号处理程序中安全使用，因为它们不是异步的信号安全。

基于`http://pubs.opengroup.org/onlinepubs/009695399/functions/pthread_sigmask.html`

```c
static sigset_t   signal_mask;  /* signals to block         */

int main (int argc, char *argv[])
{
    pthread_t sig_thr_id;      /* signal handler thread ID */
    sigemptyset (&signal_mask);
    sigaddset (&signal_mask, SIGINT);
    sigaddset (&signal_mask, SIGTERM);
    pthread_sigmask (SIG_BLOCK, &signal_mask, NULL);

    /* New threads will inherit this thread's mask */
    pthread_create (&sig_thr_id, NULL, signal_thread, NULL);

    /* APPLICATION CODE */
    ...
}

void *signal_thread (void *arg)
{
    int       sig_caught;    /* signal caught       */

    /* Use same mask as the set of signals that we'd like to know about! */
    sigwait(&signal_mask, &sig_caught);
    switch (sig_caught)
    {
    case SIGINT:     /* process SIGINT  */
        ...
        break;
    case SIGTERM:    /* process SIGTERM */
        ...
        break;
    default:         /* should normally not happen */
        fprintf (stderr, "\nUnexpected signal %d\n", sig_caught);
        break;
    }
}
```